from typing import Collection, List, Optional, Protocol

from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.typing as st


class QueryManager(Protocol):
    def storage(self) -> Storage:
        ...

    def is_compliant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: List[str],
        epsilon: Optional[float],
    ) -> bool:
        ...

    def variant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: List[str],
        epsilon: Optional[float],
    ) -> Optional[st.DataSpec]:
        ...

    def variants(self, dataspec: st.DataSpec) -> Collection[st.DataSpec]:
        ...

    def variant_constraint(
        self, dataspec: st.DataSpec
    ) -> Optional[st.VariantConstraint]:
        ...

    def transform_equivalent(
        self, transform: st.Transform, dp: bool
    ) -> Optional[st.Transform]:
        ...

    def is_pe_preserving(self, transform: st.Transform) -> bool:
        ...

    def is_differentially_private(self, transform: st.Transform) -> bool:
        ...

    def verifies(
        self,
        variant_constraint: st.VariantConstraint,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        epsilon: Optional[float],
    ) -> bool:
        ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Predicate(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    class Simple(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        OPERATOR_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        operator: typing.Text = ...
        """==, <, <=, in, etc."""

        value: typing.Text = ...
        def __init__(self,
            *,
            operator : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"operator",b"operator",u"value",b"value"]) -> None: ...

    class Inter(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        PREDICATES_FIELD_NUMBER: builtins.int
        @property
        def predicates(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Predicate]: ...
        def __init__(self,
            *,
            predicates : typing.Optional[typing.Iterable[global___Predicate]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"predicates",b"predicates"]) -> None: ...

    class Union(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        PREDICATES_FIELD_NUMBER: builtins.int
        @property
        def predicates(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Predicate]: ...
        def __init__(self,
            *,
            predicates : typing.Optional[typing.Iterable[global___Predicate]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"predicates",b"predicates"]) -> None: ...

    class Comp(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        PREDICATE_FIELD_NUMBER: builtins.int
        @property
        def predicate(self) -> global___Predicate: ...
        def __init__(self,
            *,
            predicate : typing.Optional[global___Predicate] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal[u"predicate",b"predicate"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"predicate",b"predicate"]) -> None: ...

    SIMPLE_FIELD_NUMBER: builtins.int
    INTER_FIELD_NUMBER: builtins.int
    UNION_FIELD_NUMBER: builtins.int
    COMP_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    @property
    def simple(self) -> global___Predicate.Simple: ...
    @property
    def inter(self) -> global___Predicate.Inter: ...
    @property
    def union(self) -> global___Predicate.Union: ...
    @property
    def comp(self) -> global___Predicate.Comp: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Other properties"""
        pass
    def __init__(self,
        *,
        simple : typing.Optional[global___Predicate.Simple] = ...,
        inter : typing.Optional[global___Predicate.Inter] = ...,
        union : typing.Optional[global___Predicate.Union] = ...,
        comp : typing.Optional[global___Predicate.Comp] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"comp",b"comp",u"inter",b"inter",u"predicate",b"predicate",u"simple",b"simple",u"union",b"union"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"comp",b"comp",u"inter",b"inter",u"predicate",b"predicate",u"properties",b"properties",u"simple",b"simple",u"union",b"union"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"predicate",b"predicate"]) -> typing.Optional[typing_extensions.Literal["simple","inter","union","comp"]]: ...
global___Predicate = Predicate



"""A few base implementations of basic Protocols
"""
from typing import (
    Collection,
    Mapping,
    MutableSet,
    Optional,
    Type,
    TypeVar,
    cast,
)
from uuid import UUID
import hashlib

from sarus_data_spec.context.typing import Context, HasContext
from sarus_data_spec.manager.typing import HasManager, Manager
from sarus_data_spec.protobuf.typing import Protobuf, ProtobufWithUUID
from sarus_data_spec.protobuf.utilities import copy, json, serialize, type_name
from sarus_data_spec.storage.typing import HasStorage, Storage
import sarus_data_spec.context as sc
import sarus_data_spec.typing as st

P = TypeVar('P', bound=Protobuf, covariant=True)


class Base(st.HasProtobuf[P], st.Value, st.Frozen, HasContext):
    """An object with value semantics, properties, backed by a Protobuf, with
    a default implementation"""

    def __init__(self, protobuf: P) -> None:
        self._protobuf: P = copy(protobuf)
        self._freeze()
        self._context = sc.global_context()

    def protobuf(self) -> P:
        return copy(self._protobuf)

    def prototype(self) -> Type[P]:
        raise NotImplementedError

    def type_name(self) -> str:
        return type_name(self._protobuf)

    def __bytes__(self) -> bytes:
        return serialize(self._protobuf)

    def __repr__(self) -> str:
        return json(self._protobuf)

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, value: object) -> bool:
        if isinstance(value, st.Value):
            return bytes(self) == bytes(value)
        else:
            return False

    def __hash__(self) -> int:
        return hash(bytes(self))

    def __getitem__(self, key: str) -> str:
        return cast(str, self._protobuf.properties[key])

    def properties(self) -> Mapping[str, str]:
        return self.protobuf().properties

    def _checksum(self) -> bytes:
        """Compute an md5 checksum"""
        md5 = hashlib.md5()
        md5.update(serialize(self._protobuf))
        return md5.digest()

    def _freeze(self) -> None:
        self._frozen_checksum = self._checksum()

    def _frozen(self) -> bool:
        return self._checksum() == self._frozen_checksum

    def context(self) -> Context:
        return self._context


PU = TypeVar('PU', bound=ProtobufWithUUID, covariant=True)


class Referrable(Base[PU], st.Referrable[PU], HasStorage, HasManager):
    def __init__(self, protobuf: PU) -> None:
        super().__init__(protobuf)
        # Pay attention to the fact self._freeze() in super
        # will call the implementation bellow,
        # no need to call it again although it is idempotent
        self.storage().store(self)

    def _freeze(self) -> None:
        """Referrable objects keep a uuid consistent with the checksum.
        The checksum is computed with empty uuid
        """
        self._protobuf.uuid = ''
        self._frozen_checksum = self._checksum()
        self._protobuf.uuid = UUID(bytes=self._frozen_checksum).hex

    def _frozen(self) -> bool:
        """The checksum is computed with empty uuid."""
        uuid = self._protobuf.uuid
        self._protobuf.uuid = ''
        result = (self._checksum() == self._frozen_checksum) and (
            uuid == UUID(bytes=self._frozen_checksum).hex
        )
        self._protobuf.uuid = uuid
        return result

    def uuid(self) -> str:
        return self._protobuf.uuid

    def referring(
        self, type_name: Optional[str] = None
    ) -> Collection[st.Referring]:
        return self.storage().referring(self, type_name)

    def storage(self) -> Storage:
        return self.context().storage()

    def manager(self) -> Manager:
        return self.context().manager()


class Referring(Referrable[PU], st.Referring[PU]):
    _referred: MutableSet[str] = set()

    def __init__(self, protobuf: PU) -> None:
        super().__init__(protobuf)

    def referred(self) -> Collection[st.Referrable]:
        result = set()
        for value in self._referred:
            referred = self.storage().referrable(value)
            if referred is not None:
                result.add(referred)
        return result



from sarus_data_spec.protobuf.attribute_pb2 import Attribute
from sarus_data_spec.protobuf.bounds_pb2 import Bounds
from sarus_data_spec.protobuf.constraint_pb2 import (
    ConstraintKind,
    VariantConstraint,
)
from sarus_data_spec.protobuf.dataset_pb2 import Dataset
from sarus_data_spec.protobuf.manager_pb2 import Manager
from sarus_data_spec.protobuf.marginals_pb2 import Marginals
from sarus_data_spec.protobuf.path_pb2 import Path
from sarus_data_spec.protobuf.predicate_pb2 import Predicate
from sarus_data_spec.protobuf.relation_pb2 import Relation
from sarus_data_spec.protobuf.scalar_pb2 import Scalar
from sarus_data_spec.protobuf.schema_pb2 import Schema
from sarus_data_spec.protobuf.size_pb2 import Size
from sarus_data_spec.protobuf.statistics_pb2 import Distribution, Statistics
from sarus_data_spec.protobuf.status_pb2 import Status
from sarus_data_spec.protobuf.transform_pb2 import Transform
from sarus_data_spec.protobuf.type_pb2 import Type
from sarus_data_spec.protobuf.utilities import (
    copy,
    dejson,
    deserialize,
    dict_deserialize,
    dict_serialize,
    json,
    json_deserialize,
    json_serialize,
    message,
    message_type,
    serialize,
    type_name,
    unwrap,
    wrap,
)



from __future__ import annotations

from typing import Dict, Type

from sarus_data_spec.base import Referring
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Attribute(Referring[sp.Attribute]):
    """A python class to describe attributes"""

    def __init__(self, protobuf: sp.Attribute) -> None:
        self._referred = {protobuf.object}
        # This has to be defined before it is initialized
        super().__init__(protobuf=protobuf)

    def prototype(self) -> Type[sp.Attribute]:
        """Return the type of the underlying protobuf."""
        return sp.Attribute


def attach_properties(
    node: st.DataSpec, properties: Dict[str, str]
) -> Attribute:
    return Attribute(
        sp.Attribute(
            object=node.uuid(),
            properties=properties,
        )
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/status.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/status.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%sarus_data_spec/protobuf/status.proto\x12\x0fsarus_data_spec\"\xe5\x05\n\x06Status\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61taspec\x18\x02 \x01(\t\x12\x0f\n\x07manager\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tetime\x18\x04 \x01(\t\x12<\n\x0btask_stages\x18\x05 \x03(\x0b\x32\'.sarus_data_spec.Status.TaskStagesEntry\x12;\n\nproperties\x18\x06 \x03(\x0b\x32\'.sarus_data_spec.Status.PropertiesEntry\x1aP\n\x0fTaskStagesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.sarus_data_spec.Status.Stage:\x02\x38\x01\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x97\x03\n\x05Stage\x12\x38\n\x07pending\x18\x01 \x01(\x0b\x32%.sarus_data_spec.Status.Stage.PendingH\x00\x12>\n\nprocessing\x18\x02 \x01(\x0b\x32(.sarus_data_spec.Status.Stage.ProcessingH\x00\x12\x34\n\x05ready\x18\x03 \x01(\x0b\x32#.sarus_data_spec.Status.Stage.ReadyH\x00\x12\x34\n\x05\x65rror\x18\x04 \x01(\x0b\x32#.sarus_data_spec.Status.Stage.ErrorH\x00\x12\x41\n\nproperties\x18\x05 \x03(\x0b\x32-.sarus_data_spec.Status.Stage.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\t\n\x07Pending\x1a\x0c\n\nProcessing\x1a\x07\n\x05Ready\x1a\x07\n\x05\x45rrorB\x07\n\x05stageb\x06proto3'
)




_STATUS_TASKSTAGESENTRY = _descriptor.Descriptor(
  name='TaskStagesEntry',
  full_name='sarus_data_spec.Status.TaskStagesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Status.TaskStagesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Status.TaskStagesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=339,
)

_STATUS_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Status.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Status.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Status.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=341,
  serialized_end=390,
)

_STATUS_STAGE_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Status.Stage.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Status.Stage.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Status.Stage.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=341,
  serialized_end=390,
)

_STATUS_STAGE_PENDING = _descriptor.Descriptor(
  name='Pending',
  full_name='sarus_data_spec.Status.Stage.Pending',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=750,
  serialized_end=759,
)

_STATUS_STAGE_PROCESSING = _descriptor.Descriptor(
  name='Processing',
  full_name='sarus_data_spec.Status.Stage.Processing',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=761,
  serialized_end=773,
)

_STATUS_STAGE_READY = _descriptor.Descriptor(
  name='Ready',
  full_name='sarus_data_spec.Status.Stage.Ready',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=775,
  serialized_end=782,
)

_STATUS_STAGE_ERROR = _descriptor.Descriptor(
  name='Error',
  full_name='sarus_data_spec.Status.Stage.Error',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=784,
  serialized_end=791,
)

_STATUS_STAGE = _descriptor.Descriptor(
  name='Stage',
  full_name='sarus_data_spec.Status.Stage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pending', full_name='sarus_data_spec.Status.Stage.pending', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='processing', full_name='sarus_data_spec.Status.Stage.processing', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ready', full_name='sarus_data_spec.Status.Stage.ready', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='sarus_data_spec.Status.Stage.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Status.Stage.properties', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_STATUS_STAGE_PROPERTIESENTRY, _STATUS_STAGE_PENDING, _STATUS_STAGE_PROCESSING, _STATUS_STAGE_READY, _STATUS_STAGE_ERROR, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='stage', full_name='sarus_data_spec.Status.Stage.stage',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=393,
  serialized_end=800,
)

_STATUS = _descriptor.Descriptor(
  name='Status',
  full_name='sarus_data_spec.Status',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='uuid', full_name='sarus_data_spec.Status.uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataspec', full_name='sarus_data_spec.Status.dataspec', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='manager', full_name='sarus_data_spec.Status.manager', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='datetime', full_name='sarus_data_spec.Status.datetime', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='task_stages', full_name='sarus_data_spec.Status.task_stages', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Status.properties', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_STATUS_TASKSTAGESENTRY, _STATUS_PROPERTIESENTRY, _STATUS_STAGE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=59,
  serialized_end=800,
)

_STATUS_TASKSTAGESENTRY.fields_by_name['value'].message_type = _STATUS_STAGE
_STATUS_TASKSTAGESENTRY.containing_type = _STATUS
_STATUS_PROPERTIESENTRY.containing_type = _STATUS
_STATUS_STAGE_PROPERTIESENTRY.containing_type = _STATUS_STAGE
_STATUS_STAGE_PENDING.containing_type = _STATUS_STAGE
_STATUS_STAGE_PROCESSING.containing_type = _STATUS_STAGE
_STATUS_STAGE_READY.containing_type = _STATUS_STAGE
_STATUS_STAGE_ERROR.containing_type = _STATUS_STAGE
_STATUS_STAGE.fields_by_name['pending'].message_type = _STATUS_STAGE_PENDING
_STATUS_STAGE.fields_by_name['processing'].message_type = _STATUS_STAGE_PROCESSING
_STATUS_STAGE.fields_by_name['ready'].message_type = _STATUS_STAGE_READY
_STATUS_STAGE.fields_by_name['error'].message_type = _STATUS_STAGE_ERROR
_STATUS_STAGE.fields_by_name['properties'].message_type = _STATUS_STAGE_PROPERTIESENTRY
_STATUS_STAGE.containing_type = _STATUS
_STATUS_STAGE.oneofs_by_name['stage'].fields.append(
  _STATUS_STAGE.fields_by_name['pending'])
_STATUS_STAGE.fields_by_name['pending'].containing_oneof = _STATUS_STAGE.oneofs_by_name['stage']
_STATUS_STAGE.oneofs_by_name['stage'].fields.append(
  _STATUS_STAGE.fields_by_name['processing'])
_STATUS_STAGE.fields_by_name['processing'].containing_oneof = _STATUS_STAGE.oneofs_by_name['stage']
_STATUS_STAGE.oneofs_by_name['stage'].fields.append(
  _STATUS_STAGE.fields_by_name['ready'])
_STATUS_STAGE.fields_by_name['ready'].containing_oneof = _STATUS_STAGE.oneofs_by_name['stage']
_STATUS_STAGE.oneofs_by_name['stage'].fields.append(
  _STATUS_STAGE.fields_by_name['error'])
_STATUS_STAGE.fields_by_name['error'].containing_oneof = _STATUS_STAGE.oneofs_by_name['stage']
_STATUS.fields_by_name['task_stages'].message_type = _STATUS_TASKSTAGESENTRY
_STATUS.fields_by_name['properties'].message_type = _STATUS_PROPERTIESENTRY
DESCRIPTOR.message_types_by_name['Status'] = _STATUS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Status = _reflection.GeneratedProtocolMessageType('Status', (_message.Message,), {

  'TaskStagesEntry' : _reflection.GeneratedProtocolMessageType('TaskStagesEntry', (_message.Message,), {
    'DESCRIPTOR' : _STATUS_TASKSTAGESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.status_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.TaskStagesEntry)
    })
  ,

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _STATUS_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.status_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.PropertiesEntry)
    })
  ,

  'Stage' : _reflection.GeneratedProtocolMessageType('Stage', (_message.Message,), {

    'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
      'DESCRIPTOR' : _STATUS_STAGE_PROPERTIESENTRY,
      '__module__' : 'sarus_data_spec.protobuf.status_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.Stage.PropertiesEntry)
      })
    ,

    'Pending' : _reflection.GeneratedProtocolMessageType('Pending', (_message.Message,), {
      'DESCRIPTOR' : _STATUS_STAGE_PENDING,
      '__module__' : 'sarus_data_spec.protobuf.status_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.Stage.Pending)
      })
    ,

    'Processing' : _reflection.GeneratedProtocolMessageType('Processing', (_message.Message,), {
      'DESCRIPTOR' : _STATUS_STAGE_PROCESSING,
      '__module__' : 'sarus_data_spec.protobuf.status_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.Stage.Processing)
      })
    ,

    'Ready' : _reflection.GeneratedProtocolMessageType('Ready', (_message.Message,), {
      'DESCRIPTOR' : _STATUS_STAGE_READY,
      '__module__' : 'sarus_data_spec.protobuf.status_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.Stage.Ready)
      })
    ,

    'Error' : _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), {
      'DESCRIPTOR' : _STATUS_STAGE_ERROR,
      '__module__' : 'sarus_data_spec.protobuf.status_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.Stage.Error)
      })
    ,
    'DESCRIPTOR' : _STATUS_STAGE,
    '__module__' : 'sarus_data_spec.protobuf.status_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Status.Stage)
    })
  ,
  'DESCRIPTOR' : _STATUS,
  '__module__' : 'sarus_data_spec.protobuf.status_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.Status)
  })
_sym_db.RegisterMessage(Status)
_sym_db.RegisterMessage(Status.TaskStagesEntry)
_sym_db.RegisterMessage(Status.PropertiesEntry)
_sym_db.RegisterMessage(Status.Stage)
_sym_db.RegisterMessage(Status.Stage.PropertiesEntry)
_sym_db.RegisterMessage(Status.Stage.Pending)
_sym_db.RegisterMessage(Status.Stage.Processing)
_sym_db.RegisterMessage(Status.Stage.Ready)
_sym_db.RegisterMessage(Status.Stage.Error)


_STATUS_TASKSTAGESENTRY._options = None
_STATUS_PROPERTIESENTRY._options = None
_STATUS_STAGE_PROPERTIESENTRY._options = None
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class ConstraintKind(_ConstraintKind, metaclass=_ConstraintKindEnumTypeWrapper):
    pass
class _ConstraintKind:
    V = typing.NewType('V', builtins.int)
class _ConstraintKindEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ConstraintKind.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    SYNTHETIC = ConstraintKind.V(0)
    DP = ConstraintKind.V(1)
    PUBLIC = ConstraintKind.V(2)
    PEP = ConstraintKind.V(3)

SYNTHETIC = ConstraintKind.V(0)
DP = ConstraintKind.V(1)
PUBLIC = ConstraintKind.V(2)
PEP = ConstraintKind.V(3)
global___ConstraintKind = ConstraintKind


class VariantConstraint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"value",b"value"]) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    DATASPEC_FIELD_NUMBER: builtins.int
    CONSTRAINT_KIND_FIELD_NUMBER: builtins.int
    REQUIRED_CONTEXT_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    dataspec: typing.Text = ...
    constraint_kind: global___ConstraintKind.V = ...
    @property
    def required_context(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        dataspec : typing.Text = ...,
        constraint_kind : global___ConstraintKind.V = ...,
        required_context : typing.Optional[typing.Iterable[typing.Text]] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"constraint_kind",b"constraint_kind",u"dataspec",b"dataspec",u"properties",b"properties",u"required_context",b"required_context",u"uuid",b"uuid"]) -> None: ...
global___VariantConstraint = VariantConstraint



from typing import Any, AsyncIterator
import pickle as pkl

import pandas as pd
import pyarrow as pa

from sarus_data_spec.manager.asyncio.utils import async_iter
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

from .numpy import np_mean, np_std
from .pandas import (
    pd_abs,
    pd_agg,
    pd_any,
    pd_apply,
    pd_concat,
    pd_corr,
    pd_count,
    pd_describe,
    pd_drop,
    pd_droplevel,
    pd_eq,
    pd_fillna,
    pd_get_dummies,
    pd_join,
    pd_kurtosis,
    pd_loc,
    pd_mad,
    pd_mean,
    pd_median,
    pd_quantile,
    pd_rename,
    pd_round,
    pd_select_dtypes,
    pd_skew,
    pd_sort_values,
    pd_std,
    pd_sum,
    pd_to_dict,
    pd_transpose,
    pd_unique,
    pd_value_counts,
)
from .pandas_profiling import pd_profile_report
from .sklearn import (
    sk_accuracy_score,
    sk_adaboost_classifier,
    sk_adaboost_regressor,
    sk_affinity_propagation,
    sk_agglomerative_clustering,
    sk_average_precision_score,
    sk_bagging_classifier,
    sk_bagging_regressor,
    sk_birch,
    sk_classification_report,
    sk_confusion_matrix,
    sk_cross_val_score,
    sk_dbscan,
    sk_extra_trees_classifier,
    sk_extra_trees_regressor,
    sk_f1_score,
    sk_feature_agglomeration,
    sk_fit,
    sk_gradient_boosting_classifier,
    sk_gradient_boosting_regressor,
    sk_hist_gradient_boosting_classifier,
    sk_hist_gradient_boosting_regressor,
    sk_isolation_forest,
    sk_kmeans,
    sk_label_encoder_fit_transform,
    sk_mean_shift,
    sk_minibatch_kmeans,
    sk_onehot,
    sk_optics,
    sk_pca,
    sk_plot_confusion_matrix,
    sk_precision_recall_curve,
    sk_precision_score,
    sk_random_forest_classifier,
    sk_random_forest_classifier_predict,
    sk_random_forest_regressor,
    sk_random_trees_embedding,
    sk_recall_score,
    sk_repeated_stratified_kfold,
    sk_roc_auc_score,
    sk_roc_curve,
    sk_scale,
    sk_spectral_biclustering,
    sk_spectral_clustering,
    sk_spectral_coclustering,
    sk_stacking_classifier,
    sk_stacking_regressor,
    sk_train_test_split,
    sk_voting_classifier,
    sk_voting_regressor,
)
from .std import (
    _abs,
    _and,
    _or,
    _round,
    add,
    div,
    getitem,
    greater_equal,
    greater_than,
    invert,
    length,
    lower_equal,
    lower_than,
    modulo,
    mul,
    neg,
    not_equal,
    pos,
    sub,
)
from .xgboost import xgb_classifier, xgb_classifier_predict


async def arrow_external(
    dataset: st.Dataset, batch_size: int
) -> AsyncIterator[pa.RecordBatch]:
    """Call external and convert the result to a RecordBatch iterator.

    We consider that external ops results are Datasets. For now, we consider
    that pandas.DataFrame are Datasets. For instance, the pd.loc operation only
    selects a subset of a Dataset and so is a Dataset.

    We call the implementation of `external` which returns arbitrary values,
    check that the result is indeed a DataFrame and convert it to a RecordBatch
    iterator.
    """
    val = await external(dataset)
    if isinstance(val, pd.DataFrame):
        return async_iter(
            pa.Table.from_pandas(val).to_batches(max_chunksize=batch_size)
        )

    else:
        raise TypeError(f"Cannot convert {type(val)} to Arrow batches.")


async def external(dataspec: st.DataSpec) -> Any:
    """Route an externally transformed Dataspec to its implementation."""
    transform_spec = dataspec.transform().protobuf().spec
    external_op = sp.Transform.ExternalOp.Name(transform_spec.external.op)
    implemented_ops = {
        "ADD": add,
        "MUL": mul,
        "SUB": sub,
        "DIV": div,
        "INVERT": invert,
        "GETITEM": getitem,
        "LEN": length,
        "GT": greater_than,
        "GE": greater_equal,
        "LT": lower_than,
        "LE": lower_equal,
        "NE": not_equal,
        "MOD": modulo,
        "ROUND": _round,
        "AND": _and,
        "OR": _or,
        "ABS": _abs,
        "POS": pos,
        "NEG": neg,
        "PD_LOC": pd_loc,
        "PD_EQ": pd_eq,
        "PD_MEAN": pd_mean,
        "PD_STD": pd_std,
        "PD_ANY": pd_any,
        "PD_DESCRIBE": pd_describe,
        "PD_SELECT_DTYPES": pd_select_dtypes,
        "NP_MEAN": np_mean,
        "NP_STD": np_std,
        "PD_PROFILE_REPORT": pd_profile_report,
        "SK_FIT": sk_fit,
        "SK_SCALE": sk_scale,
        'PD_QUANTILE': pd_quantile,
        'PD_SUM': pd_sum,
        'PD_FILLNA': pd_fillna,
        'PD_ROUND': pd_round,
        'PD_RENAME': pd_rename,
        'PD_COUNT': pd_count,
        'PD_TRANSPOSE': pd_transpose,
        'PD_UNIQUE': pd_unique,
        'PD_VALUE_COUNTS': pd_value_counts,
        'PD_TO_DICT': pd_to_dict,
        'PD_APPLY': pd_apply,
        'PD_MEDIAN': pd_median,
        'PD_ABS': pd_abs,
        'PD_MAD': pd_mad,
        'PD_SKEW': pd_skew,
        'PD_KURTOSIS': pd_kurtosis,
        'PD_AGG': pd_agg,
        'PD_DROPLEVEL': pd_droplevel,
        'PD_SORT_VALUES': pd_sort_values,
        'PD_DROP': pd_drop,
        'PD_CORR': pd_corr,
        "SK_ONEHOT": sk_onehot,
        "SK_PCA": sk_pca,
        # cluster
        "SK_AFFINITY_PROPAGATION": sk_affinity_propagation,
        "SK_AGGLOMERATIVE_CLUSTERING": sk_agglomerative_clustering,
        "SK_BIRCH": sk_birch,
        "SK_DBSCAN": sk_dbscan,
        "SK_FEATURE_AGGLOMERATION": sk_feature_agglomeration,
        "SK_KMEANS": sk_kmeans,
        "SK_MINIBATCH_KMEANS": sk_minibatch_kmeans,
        "SK_MEAN_SHIFT": sk_mean_shift,
        "SK_OPTICS": sk_optics,
        "SK_SPECTRAL_CLUSTERING": sk_spectral_clustering,
        "SK_SPECTRAL_BICLUSTERING": sk_spectral_biclustering,
        "SK_SPECTRAL_COCLUSTERING": sk_spectral_coclustering,
        # ensemble
        "SK_ADABOOST_CLASSIFIER": sk_adaboost_classifier,
        "SK_ADABOOST_REGRESSOR": sk_adaboost_regressor,
        "SK_BAGGING_CLASSIFIER": sk_bagging_classifier,
        "SK_BAGGING_REGRESSOR": sk_bagging_regressor,
        "SK_EXTRA_TREES_CLASSIFIER": sk_extra_trees_classifier,
        "SK_EXTRA_TREES_REGRESSOR": sk_extra_trees_regressor,
        "SK_GRADIENT_BOOSTING_CLASSIFIER": sk_gradient_boosting_classifier,
        "SK_GRADIENT_BOOSTING_REGRESSOR": sk_gradient_boosting_regressor,
        "SK_ISOLATION_FOREST": sk_isolation_forest,
        "SK_RANDOM_FOREST_CLASSIFIER": sk_random_forest_classifier,
        "SK_RANDOM_FOREST_REGRESSOR": sk_random_forest_regressor,
        "SK_RANDOM_TREES_EMBEDDING": sk_random_trees_embedding,
        "SK_STACKING_CLASSIFIER": sk_stacking_classifier,
        "SK_STACKING_REGRESSOR": sk_stacking_regressor,
        "SK_VOTING_CLASSIFIER": sk_voting_classifier,
        "SK_VOTING_REGRESSOR": sk_voting_regressor,
        "SK_HIST_GRADIENT_BOOSTING_CLASSIFIER": sk_hist_gradient_boosting_classifier,
        "SK_HIST_GRADIENT_BOOSTING_REGRESSOR": sk_hist_gradient_boosting_regressor,
        # model selection
        "SK_CROSS_VAL_SCORE": sk_cross_val_score,
        "SK_TRAIN_TEST_SPLIT": sk_train_test_split,
        "SK_REPEATED_STRATIFIED_KFOLD": sk_repeated_stratified_kfold,
        # xgb
        "XGB_CLASSIFIER": xgb_classifier,
        "XGB_CLASSIFIER_PREDICT": xgb_classifier_predict,
        # pandas 2
        "PD_GET_DUMMIES": pd_get_dummies,
        "PD_JOIN": pd_join,
        "PD_CONCAT": pd_concat,
        # metrics
        "SK_ACCURACY_SCORE": sk_accuracy_score,
        "SK_AVERAGE_PRECISION_SCORE": sk_average_precision_score,
        "SK_CLASSIFICATION_REPORT": sk_classification_report,
        "SK_CONFUSION_MATRIX": sk_confusion_matrix,
        "SK_F1_SCORE": sk_f1_score,
        "SK_PLOT_CONFUSION_MATRIX": sk_plot_confusion_matrix,
        "SK_PRECISION_RECALL_CURVE": sk_precision_recall_curve,
        "SK_PRECISION_SCORE": sk_precision_score,
        "SK_RECALL_SCORE": sk_recall_score,
        "SK_ROC_AUC_SCORE": sk_roc_auc_score,
        "SK_ROC_CURVE": sk_roc_curve,
        # ensemble predict
        "SK_RANDOM_FOREST_CLASSIFIER_PREDICT": sk_random_forest_classifier_predict,
        "SK_LABEL_ENCODER_FIT_TRANSFORM": sk_label_encoder_fit_transform,
    }
    if external_op not in implemented_ops:
        raise NotImplementedError(
            f"{external_op} not in {list(implemented_ops.keys())}"
        )

    args = pkl.loads(transform_spec.external.arguments)
    kwargs = pkl.loads(transform_spec.external.named_arguments)
    func = implemented_ops[external_op]
    return await func(dataspec, *args, **kwargs)  # type: ignore



from typing import AsyncIterator, Collection, Iterator, TypeVar
import asyncio

T = TypeVar("T")


def iter_over_async(async_iter: AsyncIterator[T]) -> Iterator[T]:
    """Synchronous generator over an asynchronous iterator."""
    while True:
        try:
            # async_iter.__anext__() is an awaitable
            next_val = asyncio.run(async_iter.__anext__())
        except StopAsyncIteration:
            break
        else:
            yield next_val


async def async_iter(data_list: Collection[T]) -> AsyncIterator[T]:
    """Convert a collection into an AsyncIterator."""
    for data in data_list:
        yield data


async def decoupled_async_iter(
    source: AsyncIterator[T], buffer_size: int = 1
) -> AsyncIterator[T]:
    """Create a consumer/producer pattern using an asyncio.Queue."""
    queue: asyncio.Queue = asyncio.Queue(maxsize=buffer_size)

    async def producer() -> None:
        async for x in source:
            await queue.put(x)
        await queue.put(None)  # producer finished

    # Launch the iteration of source iterator
    loop = asyncio.get_running_loop()
    loop.create_task(producer())

    while True:
        x = await queue.get()
        if x is None:
            queue.task_done()
            break
        queue.task_done()
        yield x



from __future__ import annotations

from datetime import datetime
from typing import Callable, List, Mapping, Optional, Type, cast
import typing as t

from sarus_data_spec.base import Base, Referring
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.typing import Manager
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Status(Referring[sp.Status]):
    """A python class to describe status"""

    def __init__(self, protobuf: sp.Status) -> None:
        self._referred = {
            protobuf.dataspec,
            protobuf.manager,
        }  # This has to be defined before it is initialized
        super().__init__(protobuf)

    def prototype(self) -> Type[sp.Status]:
        """Return the type of the underlying protobuf."""
        return sp.Status

    def datetime(self) -> datetime:
        return datetime.fromisoformat(self.protobuf().datetime)

    def update(
        self,
        task_stages: Optional[Mapping[str, st.Stage]] = None,
        properties: Optional[Mapping[str, str]] = None,
    ) -> Status:
        # We copy the content of the current status
        proto = self.protobuf()
        # We update its timestamp and properties
        proto.datetime = datetime.now().isoformat()
        if properties is not None:
            proto.properties.update(properties)
            # Protobuf implementation of the merge
            # does not work as expected on maps...
        if task_stages is not None:
            for task in task_stages:
                stage = task_stages[task].protobuf()
                if task in proto.task_stages and proto.task_stages[
                    task
                ].WhichOneof('stage') == task_stages[
                    task
                ].protobuf().WhichOneof(
                    'stage'
                ):
                    proto.task_stages[task].properties.update(stage.properties)
                else:
                    proto.task_stages[task].CopyFrom(stage)
        # And create a new data spec object
        return Status(proto)

    def task(self, task: str) -> t.Optional[Stage]:
        stage = self.protobuf().task_stages.get(task, None)
        return Stage(stage) if stage is not None else stage

    def pending(
        self,
    ) -> bool:
        """this is true if all tasks have status at least pending
        (i.e. pending, processing or ready) and at least one is pending"""
        has_one = False
        all_better = True
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                has_one = has_one or task_stages[task].HasField('pending')
                all_better = all_better and not (
                    task_stages[task].HasField('error')
                )
        return has_one and all_better

    def processing(
        self,
    ) -> bool:
        """this is true if all tasks have status at least processing
        (i.e. processing or ready) and at least one is processing"""
        has_one = False
        all_better = True
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                has_one = has_one or task_stages[task].HasField('processing')
                all_better = all_better and not (
                    task_stages[task].HasField('error')
                    or task_stages[task].HasField('pending')
                )
        return has_one and all_better

    def ready(
        self,
    ) -> bool:  # TODO this should be true if all tasks have status ready
        """this is true if all tasks have status ready"""
        all_ready = True
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                all_ready = all_ready and task_stages[task].HasField('ready')
        return all_ready

    def error(
        self,
    ) -> bool:  # TODO this should be true if any task have status error
        """this is true if any tasks have status error"""
        has_error = False
        task_stages = self.protobuf().task_stages
        if task_stages is not None:
            for task in task_stages:
                has_error = has_error or task_stages[task].HasField('error')
        return has_error

    def dataset(self) -> Dataset:
        dataspec = self.dataspec()
        assert isinstance(dataspec, Dataset)
        return dataspec

    def dataspec(self) -> st.DataSpec:  # TODO check this works
        return cast(
            st.DataSpec, self.storage().referrable(self._protobuf.dataspec)
        )

    def owner(
        self,
    ) -> Manager:
        # TODO: Maybe find a better name, but this was shadowing
        # the actual manager of this object.  # noqa: E501
        return cast(Manager, self.storage().referrable(self._protobuf.manager))


# Builders
def status(
    dataspec: st.DataSpec,
    task_stages: Optional[Mapping[str, st.Stage]] = None,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> st.Status:
    """A builder to ease the construction of a status
    - dataspec: is the dataspec, the status is added to
    - manager: is which manager, if None,
        the default manager of the dataspec is used
    """
    # Use the right manager
    manager = dataspec.manager() if manager is None else manager
    # If a status already exists for the dataspec
    # and manager simply create an updated verion of it
    last = last_status(dataspec)
    if last is None:
        return Status(
            sp.Status(
                dataspec=dataspec.uuid(),
                manager=manager.uuid(),
                datetime=datetime.now().isoformat(),
                task_stages=None
                if task_stages is None
                else {
                    task: task_stages[task].protobuf() for task in task_stages
                },
                properties=properties,
            )
        )
    else:
        return last.update(task_stages=task_stages, properties=properties)


def last_status(
    dataspec: st.DataSpec,
    manager: Optional[Manager] = None,
    task: t.Optional[str] = None,
) -> Optional[st.Status]:
    """Return a DataSpec's last status by sorted datetime."""
    manager = dataspec.manager() if manager is None else manager
    status = cast(
        Optional[st.Status],
        manager.storage().last_referring(
            (dataspec, manager), sp.type_name(sp.Status)
        ),
    )

    if status is None or task is None:
        return status

    if status.task(task=task) is None:
        return None

    return status


def last_statuses(
    dataspec: st.DataSpec, task: t.Optional[str] = None
) -> List[st.Status]:
    """Return a list composed by the last status of every
    DataSpec's manager."""
    managers = dataspec.storage().type_name(sp.type_name(sp.Manager))
    statuses = []
    for manager in managers:
        stt = last_status(
            dataspec,
            cast(
                Optional[Manager],
                manager,
            ),
        )
        if stt is not None:
            if task is not None:
                if stt.task(task) is not None:
                    statuses.append(stt)
            else:
                statuses.append(stt)
    return statuses


class Stage(Base[sp.Status.Stage]):
    """A simple wrapper type to simplify protobuf usage"""

    def accept(self, visitor: st.StageVisitor) -> None:
        dispatch: Callable[[], None] = {
            'pending': visitor.pending,
            'processing': visitor.processing,
            'ready': visitor.ready,
            'error': visitor.error,
        }[cast(str, self.protobuf().WhichOneof('stage'))]
        dispatch()

    def stage(self) -> str:
        return cast(str, self.protobuf().WhichOneof('stage'))

    def ready(self) -> bool:
        return self.stage() == 'ready'

    def pending(self) -> bool:
        return self.stage() == 'pending'

    def processing(self) -> bool:
        return self.stage() == 'processing'

    def error(self) -> bool:
        return self.stage() == 'error'


# Builders for stages
def pending_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(
            pending=sp.Status.Stage.Pending(), properties=properties
        )
    )


def processing_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(
            processing=sp.Status.Stage.Processing(), properties=properties
        )
    )


def ready_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(ready=sp.Status.Stage.Ready(), properties=properties)
    )


def error_stage(properties: Optional[Mapping[str, str]] = None) -> Stage:
    return Stage(
        sp.Status.Stage(error=sp.Status.Stage.Error(), properties=properties)
    )


# Builders for statuses


def properties(
    dataspec: st.DataSpec,
    properties: Optional[Mapping[str, str]],
    manager: Optional[Manager] = None,
) -> st.Status:
    return status(
        dataspec,
        task_stages=None,
        properties=properties,
        manager=manager,
    )


def pending(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> st.Status:
    return status(
        dataspec,
        task_stages={task: pending_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def processing(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> st.Status:
    return status(
        dataspec,
        task_stages={task: processing_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def ready(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> st.Status:
    return status(
        dataspec,
        task_stages={task: ready_stage(properties=properties)},
        properties=None,
        manager=manager,
    )


def error(
    dataspec: st.DataSpec,
    task: str,
    properties: Optional[Mapping[str, str]] = None,
    manager: Optional[Manager] = None,
) -> st.Status:
    return status(
        dataspec,
        task_stages={task: error_stage(properties=properties)},
        properties=None,
        manager=manager,
    )
